{-----------------------------------------------------------------------------
 Unit Name: pbCocoTable
 Author:    Originally Pat Terry (CRTable.PAS), ported to Delphi by Pat Connors,
 Purpose:
   Table Handler
   =============

  (1) handles a symbol table for terminals, pragmas and nonterminals
  (2) handles a table for character classes (for scanner generation)
  (3) handles a top-down graph for productions
  (4) computes various sets (start symbols, followers, any sets)
  (5) contains procedures for grammar tests
-----------------------------------------------------------------------------}


unit pbCocoTable;

interface

uses
  Windows, Messages, SysUtils, Classes,
  pbCocoScanner, pbCocoSets, pbStrLib,
  pbCocoConstants;


type
  //TsName = string[MaxNameLength];
  TsName = string;
  TPosition = record (* position of stretch of source text *)
    beg: Integer; (* start relative to beginning of file *)
    len: Integer; (* length *)
    col: Integer; (* column number of start position *)
  end;

  //  TSymbolNodeType = (cnNonterminalSymbol, cnTerminalSymbol, cnPragma, cnWeakTerminalSymbol, snUnknown);

  TCocoNodeType = Integer;
const
  cnUnknown = 0;
  cnTerminalSymbol = 1;
  cnPragma = 2;
  cnNonterminalSymbol = 3;
  cnCharClass = 4;
  cnSingleChar = 5;
  cnWeakTerminalSymbol = 6;
  cnSymbolANY = 7;
  cnEmptyAlternative = 8;
  cnSymbolSYNC = 9;
  cnSemanticAction = 10;
  cnAlternative = 11;
  cnIteration = 12;
  cnOption = 13;
  cnNoSym = -1;
  cnEndOfFile = 0;

type
  TSymbolNode = record (* node of symbol table *)
    //typ: integer; (* nt, cnterminalSymbol, cnPragma, cnunknown *)
    snType: TCocoNodeType;
    Name, (* symbol name *)
    constant: TsName; (* named constant of symbol *)
    { TODO : What is struct? Should it be a TSymbolNodeType? }
    struct: Integer; (* typ = nt: index of FirstChars node of syntax graph *)

    (* typ = cnterminalSymbol: token kind: literal, class, ... *)
    deletable: boolean; (* typ = nt: TRUE, if nonterminal is deletable *)
    attrPos: TPosition; (* position of attributes in source text *)
    semPos: TPosition;
    (* typ = cnPragma: pos of cnsemanticAction action in source text *)
    (* typ = nt: pos of local decls in source text *)
    line: Integer; (* source text line number of symbol in this node *)
  end;

  //  TGraphNodeType = (cnNonterminalSymbol, cnTerminalSymbol, cnWeakTerminalSymbol, cnSingleChar, cnCharClass, cnSymbolANY, cnEmptyAlternative,
  //    cnSemanticAction, cnSymbolSYNC, cnAlternative, cnIteration, cnOption);

  TGraphNode = record (* node of top-down graph *)
    //typ: integer; (* nt,sts,wts,chart,class,cnsymbolANY,cnemptyAlternative,cnsemanticAction,cnsymbolsYNC,cnalternative,
    //                         cnIteration,cnOption*)
    gnType: TCocoNodeType;

    Next: Integer; (* to successor node *)
    (* next < 0: to successor of enclosing structure *)
    p1: Integer;
    (* typ IN {nt, cnterminalSymbol, cnweakTerminalSymbol}: index to symbol table *)
    (* typ = cnsymbolANY: index to anyset *)
    (* typ = cnsymbolsYNC: index to syncset *)
                         (* typ = cnalternative:
                                  index of FirstChars node of FirstChars alternative *)
    (* typ IN {cnIteration, cnOption}: FirstChars node in subexpression *)
    (* typ = chart: ordinal character value *)
    (* typ = class: index of character class *)
    p2: Integer; (* typ = cnalternative:
                                  index of FirstChars node of second alternative *)
    (* typ IN {chart, class}: transition code *)
    pos: TPosition; (* typ IN {nt, cnterminalSymbol, cnweakTerminalSymbol}:
                                  source pos of actual attributes *)
    (* typ = cnsemanticAction: source pos of cnsemanticAction action *)
    line: Integer; (* source text line number of item in this node *)
  end;

{ The next is nasty - we really want different sizes of sets
  TCRTSet   = ARRAY [0 .. maxTerminals DIV 16 (*SetSize*) ] OF BITSET;
}   // - pc - replaced by DynArray
  TCRTSet = BITARRAY;

  { This means some hacking later on from the original Modula stuff }

  //;TMarkList = array[0..maxNodes div Set_size] of BITSET;
  TMarkList = array of BITSET;

  // new - pc
  TFirstSetRec = record
    ts: TCRTSet; (* terminal symbols *)
    ready: boolean; (* TRUE = ts is complete *)
  end;

  TFirstSets = array of TFirstSetRec;

  //  TFirstSets = array[0..maxNt] of record
  //    ts: TCRTSet; (* terminal symbols *)
  //    ready: boolean; (* TRUE = ts is complete *)
  //  end;

  TFollowSetRec = record
    ts: TCRTSet; (* terminal symbols *)
    nts: TCRTSet; (* nts whose start set is to be included in ts *)
  end;

  TFollowSets = array of TFollowSetRec;

  //  TFollowSets = array[0..maxNt] of record
  //    ts: TCRTSet; (* terminal symbols *)
  //    nts: TCRTSet; (* nts whose start set is to be included in ts *)
  //  end;

  TCharClass = record
    Name: TsName; (* class name *)
    fSyncSymbolSet: Integer (* ptr to set representing the class *)
  end;

  //TSymbolTable = array[0..maxSymbols] of TSymbolNode;
  TSymbolTable = array of TSymbolNode;
  //TClassTable = array[0..maxClasses] of TCharClass;
  TClassTable = array of TCharClass;

  //TGraphList = array[0..maxNodes] of TGraphNode;
  TGraphList = array of TGraphNode;

  //TSymbolSet = array[0..maxSetNr] of TCRTSet;
  TSymbolSet = array of TCRTSet;

  // new
  TNameTableRec = record
    Name, definition: TsName
  end;

  TNameTable = array[1..maxNames] of TNameTableRec;


type
  TpbCocoTable = class(TObject)
  private
    FignoreCase: boolean;
    FgenScanner: boolean;
    FhasUses: boolean;
    FnNodes: Integer;
    FmaxT: Integer;
    FlastNt: Integer;
    Froot: Integer;
    FmaxC: Integer;
    FfirstNt: Integer;
    FmaxP: Integer;
    FCocoScanner: TpbCocoScanner;
    FhasGlobalDecs: boolean;
    FhasPublishedDecs: boolean;
    FhasPublicDecs: boolean;
    FhasDestroyDecs: boolean;
    FhasPrivateDecs: boolean;
    FhasCreateDecs: boolean;
    FhasProtectedDecs: boolean;
    FhasScannerFrame: boolean;
    FhasParserFrame: boolean;
    FhasParserTarget: boolean;
    FhasCompilerTarget: boolean;
    FhasScannerTarget: boolean;
    FhasCompilerFrame: boolean;
    fCompilerFrame: string;
    fCompilerTarget: string;
    fParserFrame: string;
    fParserTarget: string;
    fScannerFrame: string;
    fScannerTarget: string;
    FTargetDirectory: string;
    FhasTargetDirectory: boolean;
    FhasErrorsSource: boolean;
    FErrorsSource: string;
    procedure MovePragmas;
    procedure ASCIIName(ascii: char; var asciiname: TsName);
    procedure BuildName(var oldName, newName: TsName);
    procedure CompAnySets;
    procedure CompFirstSets;
    procedure CompFollowSets;
    procedure CompSyncSets;
    //    procedure GetFirstSet(sp: integer; var s: TCRTSet);
    //    procedure GetFollowSet(sp: integer; var s: TCRTSet);
    procedure PrintSet(var f: TStringStream; s: BITARRAY; indent: Integer);
    procedure SymName(symn: TsName; var conn: TsName);
    procedure SetGenScanner(const Value: boolean);
    procedure SetIgnoreCase(const Value: boolean);
    procedure SetHasUses(const Value: boolean);
    procedure SetnNodes(const Value: Integer);
    procedure SetRoot(const Value: Integer);
    procedure SetMaxP(const Value: Integer);
    procedure SetSymNames(const Value: boolean);
    procedure SetCocoScanner(const Value: TpbCocoScanner);
    procedure pbWriteText(var f: TStringStream; s: string; w: Integer);
    procedure SethasGlobalDecs(const Value: boolean);
    procedure SethasCreateDecs(const Value: boolean);
    procedure SethasDestroyDecs(const Value: boolean);
    procedure SethasPrivateDecs(const Value: boolean);
    procedure SethasProtectedDecs(const Value: boolean);
    procedure SethasPublicDecs(const Value: boolean);
    procedure SethasPublishedDecs(const Value: boolean);
    procedure SethasCompilerFrame(const Value: boolean);
    procedure SethasCompilerTarget(const Value: boolean);
    procedure SethasParserFrame(const Value: boolean);
    procedure SethasParserTarget(const Value: boolean);
    procedure SethasScannerFrame(const Value: boolean);
    procedure SethasScannerTarget(const Value: boolean);
    function GetCompilerFrame: string;
    function GetCompilerTarget: string;
    function GetParserFrame: string;
    function GetParserTarget: string;
    function GetScannerFrame: string;
    function GetScannerTarget: string;
    procedure SetParserTarget(const Value: string);
    procedure InitDynamicArrays;
    procedure SetTargetDirectory(const Value: string);
    procedure SethasTargetDirectory(const Value: boolean);
    procedure SethasErrorsSource(const Value: boolean);
    procedure SetErrorsSource(const Value: string);
  protected
    { Protected declarations }
  public
    { Public declarations }
    fSymbolTable: TSymbolTable; (* symbol table for terminals,
                                         pragmas, and nonterminals *)
    fTopDownGraph: TGraphList; (* top-down graph *)
    tt: TNameTable; (* table of token name declarations *)
    FirstChars: TFirstSets;
    (* FirstChars[i]  = FirstChars symbols of fSymbolTable[i+firstNt] *)
    FollowingChars: TFollowSets;
    (* FollowingChars[i] = followers of fSymbolTable[i+firstNt] *)
    fCharacterClasses: TClassTable; (* character classes *)
    fSyncSymbolSet: TSymbolSet; (* fSyncSymbolSet[0] = all cnsymbolsYNC symbols *)
    maxSet: Integer; (* index of last symbol set *)
    lastName, dummyName: Integer; (* for unnamed character classes *)
    ch: char;
    ddt: array['A'..'Z'] of boolean;
    ignored: TCRTSet; (* characters ignored by the scanner *)
    (* parameter, debug and test switches *)
    useDeclPos,
    semDeclPos: TPosition; (* position of global semantic declarations *)

    // New Terms for the ATG file
    // GLOBALDECS - cnsymbolANY code after the uses clause and before the generated Class declaration
    globalDeclPos: Tposition;
    PrivateDeclPos: Tposition;
    ProtectedDeclPos: Tposition;
    PublicDeclPos: Tposition;
    PublishedDeclPos: Tposition;
    CreateDeclPos: Tposition;
    DestroyDeclPos: Tposition;

    CompilerFramePos: TPosition;
    CompilerTargetPos: TPosition;

    ParserFramePos: TPosition;
    ParserTargetPos: TPosition;

    ScannerFramePos: TPosition;
    ScannerTargetPos: TPosition;
    TargetDirectoryPos: Tposition;
    ErrorsPos: TPosition;
    //--------------------------------------------------------------------------
    FsymNames: boolean;
    constructor Create;
    destructor Destroy; override;
    procedure NewName(n: TsName; s: string);
    (* Inserts the pair (n, s) in the token symbol name table *)
    function NewSym(typ: TCocoNodeType; n: TsName; line: Integer): Integer;
    (* Generates a new symbol with type cnterminalSymbol and name n and returns its index *)
    procedure GetSym(sp: Integer; var sn: TSymbolNode);
    (* Gets symbol node with index sp in sn. *)
    procedure PutSym(sp: Integer; sn: TSymbolNode);
    (* Replaces symbol node with index sp by sn. *)
    function FindSym(n: TsName): Integer;
    (* Gets symbol index for identifier with name n. *)
    function NewSet(s: TCRTSet): Integer;
    (* Stores s as a new set and returns its index. *)
    procedure CompFirstSet(gp: Integer; var fs: TCRTSet);
    (* Computes start symbols of graph gp. *)
    procedure CompExpected(gp, sp: Integer; var exp: TCRTSet);
    (* Computes all symbols expected at location gp in graph of symbol sp. *)
    procedure CompDeletableSymbols;
    (* Marks deletable nonterminals and prints them. *)
    procedure CompSymbolSets;
    (* Collects FirstChars-sets, FollowingChars-sets, cnsymbolANY-sets, and cnsymbolsYNC- *)
    procedure PrintSymbolTable;
    (* Prints the symbol table (for tracing). *)
    procedure XRef;
    (* Produces a cross reference listing of all symbols. *)
    function NewClass(n: TsName; fSyncSymbolSet: TCRTSet): Integer;
    (* Defines a new character class and returns its index *)
    function ClassWithName(n: TsName): Integer;
    (* Searches for a class with the given name.  Returns its index or -1 *)
    function ClassWithSet(s: TCRTSet): Integer;
    (* Searches for a class with the given set. Returns its index or -1 *)
    procedure GetClass(n: Integer; var s: TCRTSet);
    (* Returns character class n *)
    procedure GetClassName(n: Integer; var Name: TsName);
    (* Returns the name of class n *)
    procedure GetSet(nr: Integer; var s: TCRTSet);
    (* Gives access to precomputed symbol sets *)
    function NewNode(typ: TCocoNodeType; p1, line: Integer): Integer;
(* Generates a new graph node with typ, p1, and source line number
   line and returns its index. *)
    procedure ClearMarkList(var m: TMarkList);
    (* Clears all elements of m *)
    function IsInMarkList(var s: TMarkList; x: Integer): boolean;
    (* Returns x IN s *)
    procedure InclMarkList(var s: TMarkList; x: Integer);
    (* s.INCL(x) *)
    procedure GetNode(gp: Integer; var n: TGraphNode);
    (* Gets graph node with index gp in n. *)
    procedure PutNode(gp: Integer; n: TGraphNode);
    (* Replaces graph node with index gp by n. *)
    procedure ConcatAlt(var gL1, gR1: Integer; gL2, gR2: Integer);
(* Makes (gL2, gR2) an alternative of the graph (gL1, gR1).
   The resulting graph is identified by (gL1, gR1). *)

    procedure ConcatSeq(var gL1, gR1: Integer; gL2, gR2: Integer);
(* Concatenates graph (gL1, gR1) with graph (gL2, gR2) via next-chain.
   The resulting graph is identified by (gL1, gR1). *)
    procedure MakeFirstAlt(var gL, gR: Integer);
    (* Generates an cnalternative-node with (gL, gR) as its FirstChars and only alternative *)
    procedure MakeIteration(var gL, gR: Integer);
(* Encloses the graph (gL, gR) into an iteration construct.
   The resulting graph is identified by (gL, gR). *)
    procedure MakeOption(var gL, gR: Integer);
(* Encloses the graph (gL, gR) into an option construct.
   The resulting graph is identified by (gL, gR). *)
    procedure CompleteGraph(gp: Integer);
    (* Lets right ends of graph gp be 0 *)
    procedure StrToGraph(s: string; var gL, gR: Integer);
    (* Generates linear graph from characters in s *)
    function DelGraph(gp: Integer): boolean;
    (* TRUE, if (sub) graph with root gp is deletable. *)
    function DelNode(gn: TGraphNode): boolean;
(* TRUE, if graph node gn is deletable, i.e. can be derived into the
   empty string. *)
    procedure PrintGraph;
    (* Prints the graph (for tracing). *)
    procedure FindCircularProductions(var ok: boolean);
(* Finds and prints the circular part of the grammar.
   ok = TRUE means no circular part. *)
    procedure LL1Test(var ll1: boolean);
(* Checks if the grammar satisfies the LL(1) conditions.
   ll1 = TRUE means no LL(1)-conflicts. *)
    procedure TestCompleteness(var ok: boolean);
    (* ok = TRUE, if all nonterminals have productions. *)
    procedure TestIfAllNtReached(var ok: boolean);
    (* ok = TRUE, if all nonterminals can be reached from the start symbol. *)
    procedure TestIfNtToTerm(var ok: boolean);
    (* ok = TRUE, if all nonterminals can be reduced to terminals. *)
    procedure AssignSymNames(default: boolean; var thereExists: boolean);
    procedure Restriction(n, limit: Integer);
    (* Signal compiler restriction and abort program *)
    function CocoNodeTypeToString(const CN: TCocoNodeType): string;
  published
    { Published declarations }
    property maxT: Integer read FmaxT;
    (* terminals stored from 0 .. maxT in symbol table *)
    property maxP: Integer read FmaxP write SetMaxP;
    (* pragmas stored from maxT+1..maxP in symbol table *)
    property firstNt: Integer read FfirstNt;
    (* index of FirstChars nt: available after CompSymbolSets *)
    property lastNt: Integer read FlastNt;
    (* index of last nt: available after CompSymbolSets *)
    property maxC: Integer read FmaxC; (* index of last character class *)
    property nNodes: Integer read FnNodes write SetnNodes;
    (* index of last top-down graph node *)
    property root: Integer read Froot write SetRoot;
    (* index of root node, filled by ATG *)
    property genScanner: boolean read FgenScanner write SetGenScanner;
    (* TRUE: a scanner shall be generated *)
    property hasUses: boolean read FhasUses write SetHasUses;
    (* TRUE: attribute grammar had USES clause *)
    property hasGlobalDecs: boolean read FhasGlobalDecs write SethasGlobalDecs;
    property hasPrivateDecs: boolean read FhasPrivateDecs write SethasPrivateDecs;
    property hasProtectedDecs: boolean read FhasProtectedDecs write SethasProtectedDecs;
    property hasPublicDecs: boolean read FhasPublicDecs write SethasPublicDecs;
    property hasPublishedDecs: boolean read FhasPublishedDecs write SethasPublishedDecs;
    property hasCreateDecs: boolean read FhasCreateDecs write SethasCreateDecs;
    property hasDestroyDecs: boolean read FhasDestroyDecs write SethasDestroyDecs;
    property hasErrorsSource: boolean read FhasErrorsSource write SethasErrorsSource;
    property hasCompilerFrame: boolean read FhasCompilerFrame write SethasCompilerFrame;
    property hasCompilerTarget: boolean read FhasCompilerTarget write SethasCompilerTarget;
    property hasParserFrame: boolean read FhasParserFrame write SethasParserFrame;
    property hasParserTarget: boolean read FhasParserTarget write SethasParserTarget;
    property hasScannerFrame: boolean read FhasScannerFrame write SethasScannerFrame;
    property hasScannerTarget: boolean read FhasScannerTarget write SethasScannerTarget;
    property CompilerFrame: string read GetCompilerFrame write fCompilerFrame;
    property CompilerTarget: string read GetCompilerTarget write fCompilerTarget;
    property ParserFrame: string read GetParserFrame write fParserFrame;
    property ParserTarget: string read GetParserTarget write SetParserTarget;
    property ScannerFrame: string read GetScannerFrame write fScannerFrame;
    property ScannerTarget: string read GetScannerTarget write fScannerTarget;
    property TargetDirectory: string read FTargetDirectory write SetTargetDirectory;
    property hasTargetDirectory: boolean read FhasTargetDirectory write SethasTargetDirectory;
    property ErrorsSource: string read FErrorsSource write SetErrorsSource;


    property ignoreCase: boolean read FignoreCase write SetIgnoreCase;
    (* TRUE: scanner treats lower case as upper case *)
    property symNames: boolean read FsymNames write SetSymNames;
    (* TRUE: symbol names have to be assigned *)
    property CocoScanner: TpbCocoScanner read FCocoScanner write SetCocoScanner;
  end;

implementation

uses pbCocoParser;



procedure TpbCocoTable.pbWriteText(var f: TStringStream; s: string; w: Integer);
var
  slen, i: integer;
begin
  slen := Length(s);
  for i := 1 to w do
    if i <= slen then pbWrite(f, s[i])
  else
    pbWrite(f, ' ');
end;

(* Restriction          Implementation restriction
----------------------------------------------------------------------*)

procedure TpbCocoTable.Restriction(n, limit: Integer);
begin
  pbWriteLn(fCocoScanner.lst, '');
  pbWriteLn(fCocoScanner.lst, 'Restriction ' + IntToStr(n));
  case n of
    1: pbWrite(fCocoScanner.lst, ' Too many graph nodes');
    2: pbWrite(fCocoScanner.lst, ' Too many symbols');
    3: pbWrite(fCocoScanner.lst, ' Too many sets');
    4: pbWrite(fCocoScanner.lst, 'Too many character classes');
    5: pbWrite(fCocoScanner.lst, 'Too many symbol sets');
    6: pbWrite(fCocoScanner.lst, 'Too many token names');
    7: pbWrite(fCocoScanner.lst, 'Too many states');
    8: pbWrite(fCocoScanner.lst, 'Semantic text buffer overflow');
    9: pbWrite(fCocoScanner.lst, 'Circular check buffer overflow');
    10: pbWrite(fCocoScanner.lst, 'Too many literal terminals');
      -1: pbWrite(fCocoScanner.lst, 'Compiler error');
  end;
  if n > 0 then pbWriteLn(fCocoScanner.lst, ' (limited to ' + IntToStr(limit) + ')');
  (* maybe we want CRX.pbWriteStatistics; *)
  raise Exception.Create('Halting due to violated restriction');
end;

(* MovePragmas          Move pragmas after terminals
----------------------------------------------------------------------*)

procedure TpbCocoTable.MovePragmas;
var
  i: integer;
begin
  if maxP > firstNt then
  begin
    i := maxSymbols - 1;
    maxP := maxT;
    while i > lastNt do
    begin
      INC(FmaxP);
      if maxP >= firstNt then Restriction(2, maxSymbols);
      fSymbolTable[maxP] := fSymbolTable[i];
      DEC(i)
    end;
  end
end;

(* ClearMarkList        Clear mark list m
----------------------------------------------------------------------*)

procedure TpbCocoTable.ClearMarkList(var m: TMarkList);
var
  i: integer;
begin
  i := 0;
  { TODO : Check the High(m) mod }
  while i < NodeSetSize do //High(m) do //maxNodes div SetSize
  begin
    m[i] := [];
    INC(i)
  end;
end;

function TpbCocoTable.IsInMarkList(var s: TMarkList; x: Integer): boolean;
begin
  IsInMarkList := x mod Setsize in s[x div Setsize]
end;

procedure TpbCocoTable.InclMarkList(var s: TMarkList; x: Integer);
begin
  s[x div Setsize] := s[x div Setsize] + [x mod Setsize]
end;

(* GetNode              Get node with index gp in n
----------------------------------------------------------------------*)

procedure TpbCocoTable.GetNode(gp: Integer; var n: TGraphNode);
begin
  n := fTopDownGraph[gp]
end;

(* PutNode              Replace node with index gp by n
----------------------------------------------------------------------*)

procedure TpbCocoTable.PutNode(gp: Integer; n: TGraphNode);
begin
  fTopDownGraph[gp] := n
end;

(* NewName              Collects a user defined token name
----------------------------------------------------------------------*)

procedure TpbCocoTable.NewName(n: TsName; s: string);
begin
  if lastName = maxNames then Restriction(6, maxNames)
  else
  begin
    INC(lastName);
    symNames := True;
    tt[lastName].Name := n;
    tt[lastName].definition := s;
  end;
end;

(* NewSym               Generate a new symbol and return its index
----------------------------------------------------------------------*)

function TpbCocoTable.NewSym(typ: TCocoNodeType; n: TsName; line: Integer): Integer;
var
  i: integer;
begin
  i := 0;
  if maxT + 1 = firstNt then Restriction(2, maxSymbols)
  else
  begin
    case typ of
      cnTerminalSymbol:
      begin
        INC(FmaxT);
        i := maxT;
      end;
      cnPragma:
      begin
        DEC(FmaxP);
        DEC(FfirstNt);
        DEC(FlastNt);
        i := maxP;
      end;
      cnNonterminalSymbol, cnUnknown:
      begin
        DEC(FfirstNt);
        i := firstNt;
      end;
    end;
    if maxT + 1 >= firstNt then Restriction(2, maxSymbols);
    fSymbolTable[i].snType := typ;
    fSymbolTable[i].Name := n;
    fSymbolTable[i].constant := ''; (* Bug fix - PDT *)
    fSymbolTable[i].struct := 0;
    fSymbolTable[i].deletable := False;
    fSymbolTable[i].attrPos.beg := -1;
    fSymbolTable[i].semPos.beg := -1;
    fSymbolTable[i].line := line;
  end;
  NewSym := i;
end;

(* GetSym               Get symbol sp in sn
----------------------------------------------------------------------*)

procedure TpbCocoTable.GetSym(sp: Integer; var sn: TSymbolNode);
begin
  sn := fSymbolTable[sp]
end;

(* PutSym               Replace symbol with index snix by sn
----------------------------------------------------------------------*)

procedure TpbCocoTable.PutSym(sp: Integer; sn: TSymbolNode);
begin
  fSymbolTable[sp] := sn
end;

(* FindSym              Find index of symbol with name n
----------------------------------------------------------------------*)

function TpbCocoTable.FindSym(n: TsName): Integer;
var
  i: integer;
begin
  i := 0;
  (*search in terminal list*)
  while (i <= maxT) and (fSymbolTable[i].Name <> n) do INC(i);
  if i <= maxT then
  begin
    FindSym := i;
    EXIT
  end;
  i := firstNt;
  (*search in nonterminal/pragma list*)
  while (i < maxSymbols) and (fSymbolTable[i].Name <> n) do INC(i);
  if i < maxSymbols then FindSym := i
  else
    FindSym := Ord(cnNoSym)
end;

(* PrintSet             Print set s
----------------------------------------------------------------------*)

procedure TpbCocoTable.PrintSet(var f: TStringStream; s: BITARRAY; indent: Integer);
const
  maxLineLen = 80;
var
  col, i, len: integer;
  empty: boolean;
  sn: TSymbolNode;
begin
  i := 0;
  col := indent;
  empty := True;
  while i <= maxT do
  begin
    if IsIn(s, i) then
    begin
      empty := False;
      GetSym(i, sn);
      len := Length(sn.Name);
      if col + len + 2 > maxLineLen then
      begin
        pbWriteLn(f, '');
        col := 1;
        while col < indent do
        begin
          pbWrite(f, #9);
          INC(col)
        end
      end;
      pbWrite(f, sn.Name + #9);
      INC(col, len + 2)
    end;
    INC(i)
  end;
  if empty then pbWrite(f, '-- empty set --');
  pbWriteLn(f, '')
end;

(* NewSet               Stores s as a new set and return its index
----------------------------------------------------------------------*)

function TpbCocoTable.NewSet(s: TCRTSet): Integer;
  (*cnsymbolANY-set computation requires not to search if s is already in set*)
begin
  INC(maxSet);
  if maxSet > maxSetNr then Restriction(3, maxSetNr);
  fSyncSymbolSet[maxSet] := s;
  NewSet := maxSet
end;

(* CompFirstSet         Compute FirstChars symbols of (sub) graph at gp
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompFirstSet(gp: Integer; var fs: TCRTSet);
var
  visited: TMarkList;

  procedure CompFirst(gp: Integer; var fs: TCRTSet);
  var
    s: TCRTSet;
    gn: TGraphNode;
    sn: TSymbolNode;
  begin
    Clear(fs);
    while (gp <> 0) and not IsInMarkList(visited, gp) do
    begin
      GetNode(gp, gn);
      InclMarkList(visited, gp);
      case gn.gnType of
        cnNonterminalSymbol:
        begin
          if FirstChars[gn.p1 - firstNt].ready then
            Unite(fs, FirstChars[gn.p1 - firstNt].ts)
          else
          begin
            GetSym(gn.p1, sn);
            CompFirst(sn.struct, s);
            Unite(fs, s);
          end;
        end;
        cnTerminalSymbol, cnWeakTerminalSymbol:
        begin
          Incl(fs, gn.p1)
        end;
        cnSymbolANY:
        begin
          Unite(fs, fSyncSymbolSet[gn.p1])
        end;
        cnAlternative, cnIteration, cnOption:
        begin
          CompFirst(gn.p1, s);
          Unite(fs, s);
          if gn.gnType = cnAlternative then
          begin
            CompFirst(gn.p2, s);
            Unite(fs, s)
          end
        end;
        else
          (* cnemptyAlternative, cnsemanticAction, cnsymbolsYNC, ind: nothing *)
      end;
      if not DelNode(gn) then EXIT;
      gp := ABS(gn.Next)
    end
  end;
begin (* ComputeFirstSet *)
  SetLength(visited, NodeSetSize);
  ClearMarkList(visited);
  CompFirst(gp, fs);

  if ddt['I'] then
  begin
    pbWriteLn(fCocoScanner.lst, '');
    pbWriteLn(fCocoScanner.lst, 'ComputeFirstSet: gp = ' + IntToStr(gp));
    PrintSet(fCocoScanner.lst, fs, 0);
  end;
end;

(* CompFirstSets        Compute FirstChars symbols of nonterminals
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompFirstSets;
var
  i: integer;
  sn: TSymbolNode;
begin
  i := firstNt;
  while i <= lastNt do
  begin
    FirstChars[i - firstNt].ready := False;
    INC(i)
  end;
  i := firstNt;
  while i <= lastNt do
  begin (* for all nonterminals *)
    GetSym(i, sn);
    CompFirstSet(sn.struct, FirstChars[i - firstNt].ts);
    FirstChars[i - firstNt].ready := True;
    INC(i)
  end;
end;

(* CompExpected     Compute symbols expected at location gp in Symbol sp
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompExpected(gp, sp: Integer; var exp: TCRTSet);
begin
  CompFirstSet(gp, exp);
  if DelGraph(gp) then Unite(exp, FollowingChars[sp - firstNt].ts)
end;

(* CompFollowSets       Get FollowingChars symbols of nonterminals
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompFollowSets;
var
  sn: TSymbolNode;
  curSy, i, size: integer;
  visited: TMarkList;

  procedure CompFol(gp: Integer);
  var
    s: TCRTSet;
    gn: TGraphNode;
  begin
    while (gp > 0) and not IsInMarkList(visited, gp) do
    begin
      GetNode(gp, gn);
      InclMarkList(visited, gp);
      if gn.gnType = cnNonterminalSymbol then
      begin
        CompFirstSet(ABS(gn.Next), s);
        Unite(FollowingChars[gn.p1 - firstNt].ts, s);
        if DelGraph(ABS(gn.Next)) then
          Incl(FollowingChars[gn.p1 - firstNt].nts, curSy - firstNt)
      end
      else if (gn.gnType = cnOption) or (gn.gnType = cnIteration) then CompFol(gn.p1)
      else if gn.gnType = cnalternative then
      begin
        CompFol(gn.p1);
        CompFol(gn.p2)
      end;
      gp := gn.Next
    end
  end;

  procedure Complete(i: Integer);
  var
    j: integer;
  begin
    if IsInMarkList(visited, i) then EXIT;
    InclMarkList(visited, i);
    j := 0;
    while j <= lastNt - firstNt do
    begin (* for all nonterminals *)
      if IsIn(FollowingChars[i].nts, j) then
      begin
        Complete(j);
        Unite(FollowingChars[i].ts, FollowingChars[j].ts);
        (* fix 1.42 *) if i = curSy then Excl(FollowingChars[i].nts, j)
      end;
      INC(j)
    end;
  end;
begin (* GetFollowSets *)
  size := (lastNt - firstNt + 2) div SetSize;
  curSy := firstNt;
  while curSy <= lastNt do
  begin
    Clear(FollowingChars[curSy - firstNt].ts);
    i := 0;
    while i <= size do
    begin
      FollowingChars[curSy - firstNt].nts[i] := [];
      INC(i)
    end;
    INC(curSy)
  end;

  SetLength(visited, NodeSetSize);
  ClearMarkList(visited);
  curSy := firstNt; (*get direct successors of nonterminals*)
  while curSy <= lastNt do
  begin
    GetSym(curSy, sn);
    CompFol(sn.struct);
    INC(curSy)
  end;

  curSy := 0; (*add indirect successors to FollowingChars.ts*)
  while curSy <= lastNt - firstNt do
  begin
    ClearMarkList(visited);
    Complete(curSy);
    INC(curSy);
  end;
end;

(* CompAnySets          Compute all cnsymbolANY-sets
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompAnySets;
var
  curSy: integer;
  sn: TSymbolNode;

  function LeadingAny(gp: Integer; var a: TGraphNode): boolean;
  var
    gn: TGraphNode;
  begin
    if gp <= 0 then
    begin
      LeadingAny := False;
      EXIT
    end;
    GetNode(gp, gn);
    if (gn.gnType = cnsymbolANY) then
    begin
      a := gn;
      LeadingAny := True
    end
    else
      LeadingAny := (gn.gnType = cnalternative) and
        (LeadingAny(gn.p1, a) or LeadingAny(gn.p2, a)) or
        ((gn.gnType = cnOption) or (gn.gnType = cnIteration)) and LeadingAny(gn.p1,
        a) or DelNode(gn) and LeadingAny(gn.Next, a)
  end;

  procedure FindAS(gp: Integer);
  var
    gn, gn2, a: TGraphNode;
    s1, s2: TCRTSet;
    p: integer;
  begin
    while gp > 0 do
    begin
      GetNode(gp, gn);
      if (gn.gnType = cnOption) or (gn.gnType = cnIteration) then
      begin
        FindAS(gn.p1);
        if LeadingAny(gn.p1, a) then
        begin
          CompExpected(ABS(gn.Next), curSy, s1);
          Differ(fSyncSymbolSet[a.p1], s1)
        end
      end
      else if gn.gnType = cnalternative then
      begin
        p := gp;
        Clear(s1);
        while p <> 0 do
        begin
          GetNode(p, gn2);
          FindAS(gn2.p1);
          if LeadingAny(gn2.p1, a) then
          begin
            CompExpected(gn2.p2, curSy, s2);
            Unite(s2, s1);
            Differ(fSyncSymbolSet[a.p1], s2)
          end
          else
          begin
            CompFirstSet(gn2.p1, s2);
            Unite(s1, s2)
          end;
          p := gn2.p2
        end
      end;
      gp := gn.Next
    end
  end;
begin
  curSy := firstNt;
  while curSy <= lastNt do
  begin
    (* for all nonterminals *)
    GetSym(curSy, sn);
    FindAS(sn.struct);
    INC(curSy)
  end
end;

(* CompSyncSets         Compute FollowingChars symbols of cnsymbolsYNC-nodes
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompSyncSets;
var
  curSy: integer;
  sn: TSymbolNode;
  visited: TMarkList;

  procedure CompSync(gp: Integer);
  var
    s: TCRTSet;
    gn: TGraphNode;
  begin
    while (gp > 0) and not IsInMarkList(visited, gp) do
    begin
      GetNode(gp, gn);
      InclMarkList(visited, gp);
      if gn.gnType = cnsymbolsYNC then
      begin
        CompExpected(ABS(gn.Next), curSy, s);
        Incl(s, Ord(cnEndOfFile));
        Unite(fSyncSymbolSet[0], s);
        gn.p1 := NewSet(s);
        PutNode(gp, gn)
      end
      else if gn.gnType = cnalternative then
      begin
        CompSync(gn.p1);
        CompSync(gn.p2)
      end
      else if (gn.gnType = cnOption) or (gn.gnType = cnIteration) then
      begin
        CompSync(gn.p1)
      end;
      gp := gn.Next
    end
  end;
begin
  curSy := firstNt;
  SetLength(visited, NodeSetSize);
  ClearMarkList(visited);
  while curSy <= lastNt do
  begin
    GetSym(curSy, sn);
    CompSync(sn.struct);
    INC(curSy);
  end
end;

(* CompDeletableSymbols Compute all deletable symbols and print them
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompDeletableSymbols;
var
  changed, none: boolean;
  i: integer;
  sn: TSymbolNode;
begin
  repeat
    changed := False;
    i := firstNt;
    while i <= lastNt do
    begin (*for all nonterminals*)
      GetSym(i, sn);
      if not sn.deletable and DelGraph(sn.struct) then
      begin
        sn.deletable := True;
        PutSym(i, sn);
        changed := True
      end;
      INC(i)
    end;
  until not changed;
  pbWrite(FCocoScanner.lst, 'Deletable symbols:');
  i := firstNt;
  none := True;
  while i <= lastNt do
  begin
    GetSym(i, sn);
    if sn.deletable then
    begin
      none := False;
      pbWriteLn(FCocoScanner.lst, '');
      pbWrite(FCocoScanner.lst, #9 + sn.Name);
    end;
    INC(i);
  end;
  if none then
    pbWrite(FCocoScanner.lst, #9 + '-- none --');
  pbWrite(FCocoScanner.lst, #13#10);
end;

(* CompSymbolSets       Get FirstChars-sets, FollowingChars-sets, and cnsymbolsYNC-set
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompSymbolSets;
var
  i: integer;
  sn: TSymbolNode;
begin
  MovePragmas;
  CompDeletableSymbols;
  CompFirstSets;
  CompFollowSets;
  CompAnySets;
  CompSyncSets;
  if ddt['F'] then
  begin
    i := firstNt;
    pbWrite(FCocoScanner.lst, 'List of FirstChars & FollowingChars symbols:' + #13#10);
    pbWrite(FCocoScanner.lst, #13#10);
    while i <= lastNt do
    begin (* for all nonterminals *)
      GetSym(i, sn);
      pbWrite(FCocoScanner.lst, sn.Name);
      pbWrite(FCocoScanner.lst, 'FirstChars:'#9);
      PrintSet(CocoScanner.lst, FirstChars[i - firstNt].ts, 10);
      pbWrite(FCocoScanner.lst, 'FollowingChars:'#9);
      PrintSet(CocoScanner.lst, FollowingChars[i - firstNt].ts, 10);
      pbWrite(FCocoScanner.lst, #13#10);
      INC(i);
    end;
    i := 0;
    pbWrite(FCocoScanner.lst, #13#10);
    pbWrite(FCocoScanner.lst, #13#10);
    pbWrite(FCocoScanner.lst, 'List of sets (cnsymbolANY, cnsymbolsYNC):'#9);

    if maxSet < 0 then

      pbWrite(FCocoScanner.lst, #9'-- none --')
    else
      pbWrite(FCocoScanner.lst, #13#10);
    while i <= maxSet do
    begin
      pbWrite(CocoScanner.lst, #9'set[' + IntToStr(i) + '] = ');

      PrintSet(CocoScanner.lst, fSyncSymbolSet[i], 16);
      INC(i)
    end;
    pbWriteLn(CocoScanner.lst, '');
  end;
end;

(* GetSet               Get set with index nr
----------------------------------------------------------------------*)

procedure TpbCocoTable.GetSet(nr: Integer; var s: TCRTSet);
begin
  s := fSyncSymbolSet[nr]
end;

(* PrintSymbolTable     Print symbol table
----------------------------------------------------------------------*)

procedure TpbCocoTable.PrintSymbolTable;
var
  i: integer;

  procedure pbWriteBool(b: boolean);
  begin
    if b then pbWrite(CocoScanner.lst, 'TRUE ')
    else
      pbWrite(CocoScanner.lst, 'FALSE');
  end;

  procedure pbWriteTyp1(typ: TCocoNodeType);
  begin
    case typ of
      cnunknown: pbWrite(CocoScanner.lst, ' cnunknown');
      cnterminalSymbol: pbWrite(CocoScanner.lst, 'cnterminalSymbol');
      cnPragma: pbWrite(CocoScanner.lst, 'cnPragma');
      cnNonterminalSymbol: pbWrite(CocoScanner.lst, 'Nonterminal Symbol');
    end;
  end;
begin (* PrintSymbolTable *)
  pbWriteLn(CocoScanner.lst, 'TSymbolTable:');
  pbWriteLn(CocoScanner.lst, '');
  pbWrite(CocoScanner.lst, 'nr'#9'definition'#9);
  if ddt['N'] or symNames then
    pbWrite(CocoScanner.lst, 'constant'#9);
  pbWriteLn(CocoScanner.lst, 'typ'#9'hasAttrs'#9'struct'#9'del'#9'line');
  pbWriteLn(CocoScanner.lst, '');
  i := 0;
  while i < maxSymbols do
  begin
    pbWrite(CocoScanner.lst, IntToStr(i));
    pbWrite(CocoScanner.lst, #9);
    pbWrite(CocoScanner.lst, fSymbolTable[i].Name);
    if ddt['N'] or symNames then
      if i <= maxT then pbWrite(CocoScanner.lst, fSymbolTable[i].constant);
    pbWrite(CocoScanner.lst, #9);
    pbWriteTyp1(fSymbolTable[i].snType);
    pbWrite(CocoScanner.lst, #9);
    pbWriteBool(fSymbolTable[i].attrPos.beg >= 0);
    pbWrite(CocoScanner.lst, #9);
    pbWrite(CocoScanner.lst, IntToStr(fSymbolTable[i].struct));
    pbWrite(CocoScanner.lst, #9);
    pbWriteBool(fSymbolTable[i].deletable);
    pbWrite(CocoScanner.lst, #9);
    pbWriteLn(CocoScanner.lst, IntToStr(fSymbolTable[i].line));
    if i = maxT then i := firstNt
    else
      INC(i);
  end;
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
end;

(* NewClass             Define a new character class
----------------------------------------------------------------------*)

function TpbCocoTable.NewClass(n: TsName; fSyncSymbolSet: TCRTSet): Integer;
begin
  INC(FmaxC);
  if maxC > maxClasses then Restriction(4, maxClasses);
  if n[1] = '#' then
  begin
    n[2] := CHR(Ord('A') + dummyName);
    INC(dummyName)
  end;
  fCharacterClasses[maxC].Name := n;
  fCharacterClasses[maxC].fSyncSymbolSet := NewSet(fSyncSymbolSet);
  NewClass := maxC
end;

(* ClassWithName        Return index of class with name n
----------------------------------------------------------------------*)

function TpbCocoTable.ClassWithName(n: TsName): Integer;
var
  i: integer;
begin
  i := maxC;
  while (i >= 0) and (fCharacterClasses[i].Name <> n) do DEC(i);
  ClassWithName := i
end;

(* ClassWithSet        Return index of class with the specified set
----------------------------------------------------------------------*)

function TpbCocoTable.ClassWithSet(s: TCRTSet): Integer;
var
  i: integer;
begin
  i := maxC;
  while (i >= 0) and not Equal(fSyncSymbolSet[fCharacterClasses[i].fSyncSymbolSet], s) do
    DEC(i);
  ClassWithSet := i
end;

(* GetClass             Return character class n
----------------------------------------------------------------------*)

procedure TpbCocoTable.GetClass(n: Integer; var s: TCRTSet);
begin
  Assert(n < maxClasses);
  GetSet(fCharacterClasses[n].fSyncSymbolSet, s);
end;

(* GetClassName         Get the name of class n
----------------------------------------------------------------------*)

procedure TpbCocoTable.GetClassName(n: Integer; var Name: TsName);
begin
  Name := fCharacterClasses[n].Name
end;

(* XRef                 Produce a cross reference listing of all symbols
----------------------------------------------------------------------*)

procedure TpbCocoTable.XRef;
const
  maxLineLen = 80;
type
  ListPtr = ^ListNode;
  ListNode = record
    Next: ListPtr;
    line: Integer;
  end;
  ListHdr = record
    Name: TsName;
    lptr: ListPtr;
  end;
var
  gn: TGraphNode;
  col, i: integer;
  l, p, q: ListPtr;
  sn: TSymbolNode;
  xList: array[0..maxSymbols] of ListHdr;
begin (* XRef *)
  if maxT <= 0 then EXIT;
  MovePragmas; (* initialize cross reference list *);
  i := 0;
  while i <= lastNt do
  begin (* for all symbols *)
    GetSym(i, sn);
    xList[i].Name := sn.Name;
    xList[i].lptr := nil;
    if i = maxP then i := firstNt
    else
      INC(i)
  end;
  (* search lines where symbol has been referenced *)
  i := 1;
  while i <= nNodes do
  begin (* for all graph nodes *)
    GetNode(i, gn);
    if (gn.gnType = cnterminalSymbol) or (gn.gnType = cnweakTerminalSymbol) or
      (gn.gnType = cnNonterminalSymbol) then
    begin
      NEW(l);
      l.Next := xList[gn.p1].lptr;
      l.line := gn.line;
      xList[gn.p1].lptr := l
    end;
    INC(i);
  end;
  (* search lines where symbol has been defined and insert in order *)
  i := 1;
  while i <= lastNt do
  begin (*for all symbols*)
    GetSym(i, sn);
    p := xList[i].lptr;
    q := nil;
    while (p <> nil) and (p.line > sn.line) do
    begin
      q := p;
      p := p.Next
    end;
    NEW(l);
    l.Next := p;
    l.line := -sn.line;
    if q <> nil then q.Next := l
    else
      xList[i].lptr := l;
    if i = maxP then i := firstNt
    else
      INC(i)
  end;
  (* print cross reference listing *)
  pbWrite(CocoScanner.lst, 'Cross reference list:');
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
  pbWrite(CocoScanner.lst, 'Terminals:');
  pbWriteLn(CocoScanner.lst, '');
  pbWrite(CocoScanner.lst, #9'0'#9'EOF');
  pbWriteLn(CocoScanner.lst, '');
  i := 1;
  while i <= lastNt do
  begin (* for all symbols *)
    if i = maxT then
    begin
      pbWriteLn(CocoScanner.lst, '');
      pbWriteLn(CocoScanner.lst, 'Pragmas:');
    end
    else
    begin
      pbWrite(CocoScanner.lst, IntToStr(i) + #9);
      pbWriteText(CocoScanner.lst, xList[i].Name, 25);
      l := xList[i].lptr;
      col := 35;
      while l <> nil do
      begin
        if col + 5 > maxLineLen then
        begin
          pbWriteLn(CocoScanner.lst, '');
          pbWrite(CocoScanner.lst, #9);
          col := 35
        end;
        if l.line = 0 then pbWrite(CocoScanner.lst, 'undef')
        else
          pbWrite(CocoScanner.lst, IntToStr(l.line));
        INC(col, 5);
        l := l.Next
      end;
      pbWriteLn(CocoScanner.lst, '');
    end;
    if i = maxP then
    begin
      pbWriteLn(CocoScanner.lst, '');
      pbWriteLn(CocoScanner.lst, 'Nonterminals:');
      i := firstNt
    end
    else
      INC(i)
  end;
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
end;

(* NewNode              Generate a new graph node and return its index gp
----------------------------------------------------------------------*)

function TpbCocoTable.NewNode(typ: TCocoNodeType; p1, line: Integer): Integer;
begin
  INC(FnNodes);

  if nNodes > maxNodes then Restriction(1, maxNodes);
  fTopDownGraph[nNodes].gnType := typ;
  fTopDownGraph[nNodes].Next := 0;
  fTopDownGraph[nNodes].p1 := p1;
  fTopDownGraph[nNodes].p2 := 0;
  fTopDownGraph[nNodes].pos.beg := -1;
  fTopDownGraph[nNodes].pos.len := 0;
  fTopDownGraph[nNodes].pos.col := 0;
  fTopDownGraph[nNodes].line := line;
  NewNode := nNodes;
end;

(* CompleteGraph        Set right ends of graph gp to 0
----------------------------------------------------------------------*)

procedure TpbCocoTable.CompleteGraph(gp: Integer);
var
  p: integer;
begin
  while gp <> 0 do
  begin
    p := fTopDownGraph[gp].Next;
    fTopDownGraph[gp].Next := 0;
    gp := p
  end
end;

(* ConcatAlt            Make (gL2, gR2) an alternative of (gL1, gR1)
----------------------------------------------------------------------*)

procedure TpbCocoTable.ConcatAlt(var gL1, gR1: Integer; gL2, gR2: Integer);
var
  p: integer;
begin
  gL2 := NewNode(cnalternative, gL2, 0);
  p := gL1;
  while fTopDownGraph[p].p2 <> 0 do p := fTopDownGraph[p].p2;
  fTopDownGraph[p].p2 := gL2;
  p := gR1;
  while fTopDownGraph[p].Next <> 0 do p := fTopDownGraph[p].Next;
  fTopDownGraph[p].Next := gR2
end;

(* ConcatSeq            Make (gL2, gR2) a successor of (gL1, gR1)
----------------------------------------------------------------------*)

procedure TpbCocoTable.ConcatSeq(var gL1, gR1: Integer; gL2, gR2: Integer);
var
  p, q: integer;
begin
  p := fTopDownGraph[gR1].Next;
  fTopDownGraph[gR1].Next := gL2; (*head node*)
  while p <> 0 do
  begin (*substructure*)
    q := fTopDownGraph[p].Next;
    fTopDownGraph[p].Next := -gL2;
    p := q
  end;
  gR1 := gR2
end;

(* MakeFirstAlt         Generate cnalternative-node with (gL,gR) as only alternative
----------------------------------------------------------------------*)

procedure TpbCocoTable.MakeFirstAlt(var gL, gR: Integer);
begin
  gL := NewNode(cnalternative, gL, 0);
  fTopDownGraph[gL].Next := gR;
  gR := gL
end;

(* MakeIteration        Enclose (gL, gR) into iteration node
----------------------------------------------------------------------*)

procedure TpbCocoTable.MakeIteration(var gL, gR: Integer);
var
  p, q: integer;
begin
  gL := NewNode(cnIteration, gL, 0);
  p := gR;
  gR := gL;
  while p <> 0 do
  begin
    q := fTopDownGraph[p].Next;
    fTopDownGraph[p].Next := -gL;
    p := q
  end
end;

(* MakeOption           Enclose (gL, gR) into option node
----------------------------------------------------------------------*)

procedure TpbCocoTable.MakeOption(var gL, gR: Integer);
begin
  gL := NewNode(cnOption, gL, 0);
  fTopDownGraph[gL].Next := gR;
  gR := gL
end;

(* StrToGraph           Generate node chain from characters in s
----------------------------------------------------------------------*)

procedure TpbCocoTable.StrToGraph(s: string; var gL, gR: Integer);
var
  i, len: integer;
begin
  gR := 0;
  i := 2;
  len := Length(s); (*strip quotes*)
  while i < len do
  begin
    fTopDownGraph[gR].Next := NewNode(cnSingleChar, Ord(s[i]), 0);
    gR := fTopDownGraph[gR].Next;
    INC(i)
  end;
  gL := fTopDownGraph[0].Next;
  fTopDownGraph[0].Next := 0
end;

(* DelGraph             Check if graph starting with index gp is deletable
----------------------------------------------------------------------*)

function TpbCocoTable.DelGraph(gp: Integer): boolean;
var
  gn: TGraphNode;
begin
  if gp = 0 then
  begin
    DelGraph := True;
    EXIT
  end; (*end of graph found*)
  GetNode(gp, gn);
  DelGraph := DelNode(gn) and DelGraph(ABS(gn.Next));
end;

(* DelNode              Check if graph node gn is deletable
----------------------------------------------------------------------*)

function TpbCocoTable.DelNode(gn: TGraphNode): boolean;
var
  sn: TSymbolNode;

  function DelAlt(gp: Integer): boolean;
  var
    gn: TGraphNode;
  begin
    if gp <= 0 then
    begin
      DelAlt := True;
      EXIT
    end; (*end of graph found*)
    GetNode(gp, gn);
    DelAlt := DelNode(gn) and DelAlt(gn.Next);
  end;
begin
  if gn.gnType = cnNonterminalSymbol then
  begin
    GetSym(gn.p1, sn);
    DelNode := sn.deletable
  end
  else if gn.gnType = cnalternative then
    DelNode := DelAlt(gn.p1) or (gn.p2 <> 0) and DelAlt(gn.p2)
  else
    DelNode := (gn.gnType = cnemptyAlternative) or (gn.gnType = cnIteration)
      or (gn.gnType = cnOption) or
      (gn.gnType = cnsemanticAction) or (gn.gnType = cnsymbolsYNC)
end;

(* PrintGraph           Print the graph
----------------------------------------------------------------------*)

procedure TpbCocoTable.PrintGraph;
var
  i: integer;

  procedure pbWriteTyp2(typ: TCocoNodeType);
  begin
    case typ of
      cnNonterminalSymbol: pbWrite(CocoScanner.lst, 'cnNonterminalSymbol'#9);
      cnterminalSymbol: pbWrite(CocoScanner.lst, 'cnterminalSymbol'#9);
      cnweakTerminalSymbol: pbWrite(CocoScanner.lst, 'cnweakTerminalSymbol'#9);
      cnsymbolANY: pbWrite(CocoScanner.lst, 'cnsymbolANY'#9);
      cnemptyAlternative: pbWrite(CocoScanner.lst, 'cnemptyAlternative'#9);
      cnsemanticAction: pbWrite(CocoScanner.lst, 'cnsemanticAction'#9);
      cnsymbolsYNC: pbWrite(CocoScanner.lst, 'cnsymbolSYNC'#9);
      cnalternative: pbWrite(CocoScanner.lst, 'cnalternative'#9);
      cnIteration: pbWrite(CocoScanner.lst, 'cnIteration'#9);
      cnOption: pbWrite(CocoScanner.lst, 'cnOption'#9);
      else
        pbWrite(CocoScanner.lst, '---'#9)
    end;
  end;
begin (* PrintGraph *)
  pbWriteLn(CocoScanner.lst, 'TGraphList:');
  pbWriteLn(CocoScanner.lst, '');
  pbWrite(CocoScanner.lst, 'nr'#9'typ'#9'next'#9'p1'#9'p2'#9'line');
  (* useful for debugging - PDT *)
  pbWrite(CocoScanner.lst, 'posbeg'#9'poslen'#9'poscol');
  (* *)
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
  i := 0;
  while i <= nNodes do
  begin
    pbWrite(CocoScanner.lst, IntToStr(i));
    pbWrite(CocoScanner.lst, '   ');
    pbWriteTyp2(fTopDownGraph[i].gnType);
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].Next));
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].p1));
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].p2));
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].line));
    (* useful for debugging - PDT *)
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].pos.beg));
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].pos.len));
    pbWrite(CocoScanner.lst, IntToStr(fTopDownGraph[i].pos.col));
    (*  *)
    pbWriteLn(CocoScanner.lst, '');
    INC(i);
  end;
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
end;

(* FindCircularProductions      Test grammar for circular derivations
----------------------------------------------------------------------*)

procedure TpbCocoTable.FindCircularProductions(var ok: boolean);
type
  ListEntry = record
    left: Integer;
    right: Integer;
    deleted: boolean;
  end;
var
  changed, onLeftSide, onRightSide: boolean;
  i, j, listLength: integer;
  list: array[0..maxList] of ListEntry;
  singles: TMarkList;
  sn: TSymbolNode;

  procedure GetSingles(gp: Integer; var singles: TMarkList);
  var
    gn: TGraphNode;
  begin
    if gp <= 0 then EXIT;
    (* end of graph found *)
    GetNode(gp, gn);
    if gn.gnType = cnNonterminalSymbol then
      if DelGraph(ABS(gn.Next)) then InclMarkList(singles, gn.p1)
      else if (gn.gnType = cnalternative) or (gn.gnType = cnIteration)
        or (gn.gnType = cnOption) then
        if DelGraph(ABS(gn.Next)) then
        begin
          GetSingles(gn.p1, singles);
          if gn.gnType = cnalternative then GetSingles(gn.p2, singles)
        end;
    if DelNode(gn) then GetSingles(gn.Next, singles)
  end;
begin (* FindCircularProductions *)
  i := firstNt;
  listLength := 0;
  SetLength(singles, NodeSetSize);
  while i <= lastNt do
  begin (* for all nonterminals i *)
    ClearMarkList(singles);
    GetSym(i, sn);
    GetSingles(sn.struct, singles);
    (* get ntcnNonterminalSymbol's j such that i-->j *)
    j := firstNt;
    while j <= lastNt do
    begin (* for all nonterminals j *)
      if IsInMarkList(singles, j) then
      begin
        list[listLength].left := i;
        list[listLength].right := j;
        list[listLength].deleted := False;
        INC(listLength);
        if listLength > maxList then Restriction(9, listLength)
      end;
      INC(j)
    end;
    INC(i)
  end;
  repeat
    i := 0;
    changed := False;
    while i < listLength do
    begin
      if not list[i].deleted then
      begin
        j := 0;
        onLeftSide := False;
        onRightSide := False;
        while j < listLength do
        begin
          if not list[j].deleted then
          begin
            if list[i].left = list[j].right then onRightSide := True;
            if list[j].left = list[i].right then onLeftSide := True
          end;
          INC(j)
        end;
        if not onRightSide or not onLeftSide then
        begin
          list[i].deleted := True;
          changed := True
        end
      end;
      INC(i)
    end
  until not changed;
  pbWrite(CocoScanner.lst, 'Circular derivations:'#9);
  i := 0;
  ok := True;
  while i < listLength do
  begin
    if not list[i].deleted then
    begin
      ok := False;
      pbWriteLn(CocoScanner.lst, '');
      pbWrite(CocoScanner.lst, #9);
      GetSym(list[i].left, sn);
      pbWrite(CocoScanner.lst, sn.Name + '-->'#9); //sn.Name: 20, ' --> ');
      GetSym(list[i].right, sn);
      pbWrite(CocoScanner.lst, sn.Name); //sn.Name: 20);
    end;
    INC(i)
  end;
  if ok then pbWrite(CocoScanner.lst, '-- none --');
  pbWriteLn(CocoScanner.lst, '');
end;

(* LL1Test              Collect terminal sets and checks LL(1) conditions
----------------------------------------------------------------------*)

procedure TpbCocoTable.LL1Test(var ll1: boolean);
var
  sn: TSymbolNode;
  curSy: integer;

  procedure LL1Error(cond, ts: Integer);
  var
    sn: TSymbolNode;
  begin
    ll1 := False;
    pbWriteLn(CocoScanner.lst, '');
    pbWrite(CocoScanner.lst, ' LL(1) error in ');
    GetSym(curSy, sn);
    pbWrite(CocoScanner.lst, sn.Name + ': ');
    if ts > 0 then
    begin
      GetSym(ts, sn);
      pbWrite(CocoScanner.lst, sn.Name + ' is ');
    end;
    case cond of
      1: pbWrite(CocoScanner.lst, 'the start of several alternatives.');
      2: pbWrite(CocoScanner.lst, 'the start & successor of a deletable structure');
      3: pbWrite(CocoScanner.lst, 'an cnsymbolANY node that matches no symbol');
    end;
  end;

  procedure Check(cond: Integer; var s1, s2: TCRTSet);
  var
    i: integer;
  begin
    i := 0;
    while i <= maxT do
    begin
      if IsIn(s1, i) and IsIn(s2, i) then LL1Error(cond, i);
      INC(i)
    end
  end;

  procedure CheckAlternatives(gp: Integer);
  var
    gn, gn1: TGraphNode;
    s1, s2: TCRTSet;
    p: integer;
  begin
    while gp > 0 do
    begin
      GetNode(gp, gn);
      if gn.gnType = cnalternative then
      begin
        p := gp;
        Clear(s1);
        while p <> 0 do
        begin (*for all alternatives*)
          GetNode(p, gn1);
          CompExpected(gn1.p1, curSy, s2);
          Check(1, s1, s2);
          Unite(s1, s2);
          CheckAlternatives(gn1.p1);
          p := gn1.p2
        end
      end
      else if (gn.gnType = cnOption) or (gn.gnType = cnIteration) then
      begin
        CompExpected(gn.p1, curSy, s1);
        CompExpected(ABS(gn.Next), curSy, s2);
        Check(2, s1, s2);
        CheckAlternatives(gn.p1)
      end
      else if gn.gnType = cnsymbolANY then
      begin
        GetSet(gn.p1, s1);
        if Empty(s1) then LL1Error(3, 0)
      end (*e.g. {cnsymbolANY} cnsymbolANY or [cnsymbolANY] cnsymbolANY*);
      gp := gn.Next
    end
  end;
begin (* LL1Test *)
  pbWrite(CocoScanner.lst, 'LL(1) conditions:');
  curSy := firstNt;
  ll1 := True;
  while curSy <= lastNt do
  begin (*for all nonterminals*)
    GetSym(curSy, sn);
    CheckAlternatives(sn.struct);
    INC(curSy)
  end;
  if ll1 then pbWrite(CocoScanner.lst, #9' --  ok  --');
  pbWriteLn(CocoScanner.lst, '');
end;

(* TestCompleteness     Test if all nonterminals have productions
----------------------------------------------------------------------*)

procedure TpbCocoTable.TestCompleteness(var ok: boolean);
var
  sp: integer;
  sn: TSymbolNode;
begin
  pbWrite(CocoScanner.lst, 'Undefined nonterminals:  ');
  sp := firstNt;
  ok := True;
  while sp <= lastNt do
  begin (*for all nonterminals*)
    GetSym(sp, sn);
    if sn.struct = 0 then
    begin
      ok := False;
      pbWriteLn(CocoScanner.lst, '');
      pbWrite(CocoScanner.lst, #9 + sn.Name);
    end;
    INC(sp)
  end;
  if ok then pbWrite(CocoScanner.lst, ' -- none --');
  pbWriteLn(CocoScanner.lst, '');
end;

(* TestIfAllNtReached   Test if all nonterminals can be reached
----------------------------------------------------------------------*)

procedure TpbCocoTable.TestIfAllNtReached(var ok: boolean);
var
  gn: TGraphNode;
  sp: integer;
  reached: TMarkList;
  sn: TSymbolNode;

  procedure MarkReachedNts(gp: Integer);
  var
    gn: TGraphNode;
    sn: TSymbolNode;
  begin
    while gp > 0 do
    begin
      GetNode(gp, gn);
      if gn.gnType = cnNonterminalSymbol then
      begin
        if not IsInMarkList(reached, gn.p1) then (*new nt reached*)
        begin
          InclMarkList(reached, gn.p1);
          GetSym(gn.p1, sn);
          MarkReachedNts(sn.struct)
        end
      end
      else if (gn.gnType = cnalternative) or (gn.gnType = cnIteration)
        or (gn.gnType = cnOption) then
      begin
        MarkReachedNts(gn.p1);
        if gn.gnType = cnalternative then MarkReachedNts(gn.p2)
      end;
      gp := gn.Next
    end
  end;
begin (* TestIfAllNtReached *)
  SetLength(reached, NodeSetSize);
  ClearMarkList(reached);
  GetNode(root, gn);
  InclMarkList(reached, gn.p1);
  GetSym(gn.p1, sn);
  MarkReachedNts(sn.struct);
  pbWrite(CocoScanner.lst, 'Unreachable nonterminals:');
  sp := firstNt;
  ok := True;
  while sp <= lastNt do
  begin (*for all nonterminals*)
    if not IsInMarkList(reached, sp) then
    begin
      ok := False;
      GetSym(sp, sn);
      pbWriteLn(CocoScanner.lst, '');
      pbWrite(CocoScanner.lst, #9 + sn.Name)
    end;
    INC(sp)
  end;
  if ok then pbWrite(CocoScanner.lst, #9' -- none --');
  pbWriteLn(CocoScanner.lst, '');
end;

(* TestIfNtToTerm   Test if all nonterminals can be derived to terminals
----------------------------------------------------------------------*)

procedure TpbCocoTable.TestIfNtToTerm(var ok: boolean);
var
  changed: boolean;
  sp: integer;
  sn: TSymbolNode;
  termList: TMarkList;

  function IsTerm(gp: Integer): boolean;
  var
    gn: TGraphNode;
  begin
    while gp > 0 do
    begin
      GetNode(gp, gn);
      if (gn.gnType = cnNonterminalSymbol) and not IsInMarkList(termList,
        gn.p1) or (gn.gnType = cnalternative) and not IsTerm(gn.p1) and
        ((gn.p2 = 0) or not IsTerm(gn.p2)) then
      begin
        IsTerm := False;
        EXIT
      end;
      gp := gn.Next
    end;
    IsTerm := True
  end;
begin (* TestIfNtToTerm *)
  SetLength(TermList, NodeSetSize);
  ClearMarkList(termList);
  repeat
    sp := firstNt;
    changed := False;
    while sp <= lastNt do
    begin
      if not IsInMarkList(termList, sp) then
      begin
        GetSym(sp, sn);
        if IsTerm(sn.struct) then
        begin
          InclMarkList(termList, sp);
          changed := True
        end
      end;
      INC(sp)
    end
  until not changed;
  pbWrite(CocoScanner.lst, 'Underivable nonterminals:');
  sp := firstNt;
  ok := True;
  while sp <= lastNt do
  begin
    if not IsInMarkList(termList, sp) then
    begin
      ok := False;
      GetSym(sp, sn);
      //      pbWriteLn(CocoScanner.lst,'');
      pbWrite(CocoScanner.lst, '     ' + sn.Name);
      pbWriteLn(CocoScanner.lst, '');
    end;
    INC(sp)
  end;
  if ok then pbWrite(CocoScanner.lst, ' -- none --');
  pbWriteLn(CocoScanner.lst, '');
end;

(* ASCIIName            Assigns the wellknown ASCII-Name in lowercase
----------------------------------------------------------------------*)

procedure TpbCocoTable.ASCIIName(ascii: char; var asciiname: TsName);
var
  N: integer;
begin
  case ascii of
    #00: asciiname := '_nul';
    #01: asciiname := '_soh';
    #02: asciiname := '_stx';
    #03: asciiname := '_etx';
    #04: asciiname := '_eot';
    #05: asciiname := '_enq';
    #06: asciiname := '_ack';
    #07: asciiname := '_bel';
    #08: asciiname := '_bs';
    #09: asciiname := '_ht';
    #10: asciiname := '_lf';
    #11: asciiname := '_vt';
    #12: asciiname := '_ff';
    #13: asciiname := '_cr';
    #14: asciiname := '_so';
    #15: asciiname := '_si';
    #16: asciiname := '_dle';
    #17: asciiname := '_dc1';
    #18: asciiname := '_dc2';
    #19: asciiname := '_dc3';
    #20: asciiname := '_dc4';
    #21: asciiname := '_nak';
    #22: asciiname := '_syn';
    #23: asciiname := '_etb';
    #24: asciiname := '_can';
    #25: asciiname := '_em';
    #26: asciiname := '_sub';
    #27: asciiname := '_esc';
    #28: asciiname := '_fs';
    #29: asciiname := '_gs';
    #30: asciiname := '_rs';
    #31: asciiname := '_us';
    ' ': asciiname := '_sp';
    '!': asciiname := '_bang';
    '"': asciiname := '_dquote';
    '#': asciiname := '_hash';
    '$': asciiname := '_dollar';
    '%': asciiname := '_percent';
    '&': asciiname := '_and';
    '''': asciiname := '_squote';
    '(': asciiname := '_lparen';
    ')': asciiname := '_rparen';
    '*': asciiname := '_star';
    '+': asciiname := '_plus';
    ',': asciiname := '_comma';
    '-': asciiname := '_minus';
    '.': asciiname := '_point';
    '/': asciiname := '_slash';
    '0': asciiname := '_zero';
    '1': asciiname := '_one';
    '2': asciiname := '_two';
    '3': asciiname := '_three';
    '4': asciiname := '_four';
    '5': asciiname := '_five';
    '6': asciiname := '_six';
    '7': asciiname := '_seven';
    '8': asciiname := '_eight';
    '9': asciiname := '_nine';
    ':': asciiname := '_colon';
    ';': asciiname := '_semicolon';
    '<': asciiname := '_less';
    '=': asciiname := '_equal';
    '>': asciiname := '_greater';
    '?': asciiname := '_query';
    '@': asciiname := '_at';
    'A'..'Z', 'a'..'z':
    begin
      asciiname := '_ ';
      asciiname[2] := ascii
    end;
    '[': asciiname := '_lbrack';
    '\': asciiname := '_backslash';
    ']': asciiname := '_rbrack';
    '^': asciiname := '_uparrow';
    '_': asciiname := '_underscore';
    '`': asciiname := '_accent';
    '{': asciiname := '_lbrace';
    '|': asciiname := '_bar';
    '}': asciiname := '_rbrace';
    '~': asciiname := '_tilde';
    #127: asciiname := '_delete';
    else
    begin
      N := Ord(ascii);
      asciiname := 'ascii  ';
      asciiname[7] := CHR(N mod 10 + Ord('0'));
      N := N div 10;
      asciiname[6] := CHR(N mod 10 + Ord('0'));
      asciiname[5] := CHR(N div 10 + Ord('0'));
    end
  end;
end;

(* BuildName            Build new Name to represent old string
----------------------------------------------------------------------*)

procedure TpbCocoTable.BuildName(var oldName, newName: TsName);
var
  ForLoop, I: integer;
  TargetIndex: integer;
  ascName: TsName;
begin
  TargetIndex := 1;
  for ForLoop := 2 to Length(oldName) - 1 do
  begin
    SetLength(newName, MaxNameLength);
    case oldName[ForLoop] of
      'A'..'Z', 'a'..'z':
      begin
        if TargetIndex <= 255 then
        begin
          newName[TargetIndex] := oldName[ForLoop];
          INC(TargetIndex);
        end;
      end;
      else
      begin
        ASCIIName(oldName[ForLoop], ascName);

        for I := 1 to Length(ascName) do
          if TargetIndex <= MaxNameLength - 3 then
          begin
            newName[TargetIndex] := ascName[I];
            INC(TargetIndex)
          end;
      end;
    end;
  end;
  SetLength(newName, TargetIndex - 1);
end;

(* SymName              Generates a new name for a symbol constant
----------------------------------------------------------------------*)

procedure TpbCocoTable.SymName(symn: TsName; var conn: TsName);
begin
  if (symn[1] = '''') or (symn[1] = '"') then if Length(symn) = 3 then
      ASCIIName(symn[2], conn)
    else
      BuildName(symn, conn)
  else
    conn := symn;
  conn := Concat(conn, 'Sym');
end;

(* AssignSymNames     Assigns the user defined or generated token names
----------------------------------------------------------------------*)

procedure TpbCocoTable.AssignSymNames(default: boolean; var thereExists: boolean);

  procedure AssignDef(var n (*is not modified*), constant: TsName);
  var
    ForLoop: integer;
  begin
    for ForLoop := 1 to lastName do
      if n = tt[ForLoop].definition then
      begin
        constant := tt[ForLoop].Name;
        thereExists := True;
        EXIT;
      end;
    if default then SymName(n, constant)
    else
      constant := ''
  end;
var
  ForLoop: integer;
begin
  thereExists := default;
  fSymbolTable[0].constant := 'EOFSYMB';
  for ForLoop := 1 to maxP do
    AssignDef(fSymbolTable[ForLoop].Name, fSymbolTable[ForLoop].constant);
  fSymbolTable[maxT].constant := 'NO_SYMB';
end;


procedure TpbCocoTable.SetGenScanner(const Value: boolean);
begin
  FgenScanner := Value;
end;

procedure TpbCocoTable.SetIgnoreCase(const Value: boolean);
begin
  FignoreCase := Value;
end;

procedure TpbCocoTable.SetHasUses(const Value: boolean);
begin
  FhasUses := Value;
end;

procedure TpbCocoTable.SetnNodes(const Value: Integer);
begin
  FnNodes := Value;
end;

procedure TpbCocoTable.SetRoot(const Value: Integer);
begin
  Froot := Value;
end;

procedure TpbCocoTable.SetMaxP(const Value: Integer);
begin
  FmaxP := Value;
end;

procedure TpbCocoTable.SetSymNames(const Value: boolean);
begin
  FsymNames := Value;
end;


procedure TpbCocoTable.SetCocoScanner(const Value: TpbCocoScanner);
begin
  FCocoScanner := Value;
end;

constructor TpbCocoTable.Create;

  procedure Clear(var s: BITARRAY);
  var
    i: integer;
  begin
    i := 0;
    while i <= Set_Limit do
    begin
      s[i] := [];
      INC(i)
    end
  end;

  procedure Incl(var s: BITARRAY; x: Integer);
  begin
    s[x div Set_Size] := s[x div Set_Size] + [x mod Set_Size]
  end;
begin
  inherited;

  ch := 'A';
  while ch <= 'Z' do
  begin
    ddt[ch] := False;
    INC(ch)
  end;
  maxSet := 0;
  InitDynamicArrays;
  Clear(fSyncSymbolSet[0]);
  Incl(fSyncSymbolSet[0], Ord(cnEndOfFile));
  ffirstNt := maxSymbols;
  fmaxP := maxSymbols;
  fmaxT := -1;
  fmaxC := -1;
  flastNt := maxP - 1;
  dummyName := 0;
  lastName := 0;
  symNames := False;
  hasUses := False;
  hasGlobalDecs := False;
  (* The dummy node fTopDownGraph[0] ensures that none of the procedures
     above have to check for 0 indices. *)

  // We no longer need to allocate memeory this way
  // - now setting these up in the InitDynArrays Method
  //  NEW(fTopDownGraph);
  //  NEW(fSymbolTable);




  nNodes := 0;
  fTopDownGraph[0].gnType := cnEndOfFile;
  fTopDownGraph[0].p1 := 0;
  fTopDownGraph[0].p2 := 0;
  fTopDownGraph[0].Next := 0;
  fTopDownGraph[0].line := 0;
  fTopDownGraph[0].pos.beg := -1;
  fTopDownGraph[0].pos.len := 0;
  fTopDownGraph[0].pos.col := 0;
end;

destructor TpbCocoTable.Destroy;
begin
  fTopDownGraph := nil;
  fSymbolTable := nil;
  inherited;
end;

procedure TpbCocoTable.SethasGlobalDecs(const Value: boolean);
begin
  FhasGlobalDecs := Value;
end;

procedure TpbCocoTable.SethasCreateDecs(const Value: boolean);
begin
  FhasCreateDecs := Value;
end;

procedure TpbCocoTable.SethasDestroyDecs(const Value: boolean);
begin
  FhasDestroyDecs := Value;
end;

procedure TpbCocoTable.SethasPrivateDecs(const Value: boolean);
begin
  FhasPrivateDecs := Value;
end;

procedure TpbCocoTable.SethasProtectedDecs(const Value: boolean);
begin
  FhasProtectedDecs := Value;
end;

procedure TpbCocoTable.SethasPublicDecs(const Value: boolean);
begin
  FhasPublicDecs := Value;
end;

procedure TpbCocoTable.SethasPublishedDecs(const Value: boolean);
begin
  FhasPublishedDecs := Value;
end;

procedure TpbCocoTable.SethasCompilerFrame(const Value: boolean);
begin
  FhasCompilerFrame := Value;
end;

procedure TpbCocoTable.SethasCompilerTarget(const Value: boolean);
begin
  FhasCompilerTarget := Value;
end;

procedure TpbCocoTable.SethasParserFrame(const Value: boolean);
begin
  FhasParserFrame := Value;
end;

procedure TpbCocoTable.SethasParserTarget(const Value: boolean);
begin
  FhasParserTarget := Value;
end;

procedure TpbCocoTable.SethasScannerFrame(const Value: boolean);
begin
  FhasScannerFrame := Value;
end;

procedure TpbCocoTable.SethasScannerTarget(const Value: boolean);
begin
  FhasScannerTarget := Value;
end;

function TpbCocoTable.GetCompilerFrame: string;
begin
  Result := Trim(fCompilerFrame);
end;

function TpbCocoTable.GetCompilerTarget: string;
begin
  Result := Trim(fCompilerTarget);
  Result := KillFinalDelim(Result, '\');
end;

function TpbCocoTable.GetParserFrame: string;
begin
  Result := Trim(fParserFrame);
end;

function TpbCocoTable.GetParserTarget: string;
begin
  Result := Trim(fParserTarget);
  Result := KillFinalDelim(Result, '\');
end;

function TpbCocoTable.GetScannerFrame: string;
begin
  Result := Trim(fScannerFrame);
end;

function TpbCocoTable.GetScannerTarget: string;
begin
  Result := Trim(fScannerTarget);
  Result := KillFinalDelim(Result, '\');
end;

procedure TpbCocoTable.SetParserTarget(const Value: string);
begin
  fParserTarget := Value;
end;

procedure TpbCocoTable.InitDynamicArrays;
begin
  SetLength(fSyncSymbolSet, maxSetNr);
  SetLength(fTopDownGraph, maxNodes);
  SetLength(fSymbolTable, maxSymbols);
  SetLength(fCharacterClasses, maxClasses);
  SetLength(FirstChars, maxNt);
  SetLength(FollowingChars, maxNt);
end;

{-----------------------------------------------------------------------------
  Procedure: TpbCocoTable.CocoNodeTypeToString
  Purely for debugging if needed
-----------------------------------------------------------------------------}
function TpbCocoTable.CocoNodeTypeToString(const CN: TCocoNodeType): string;
begin
  //  case CN of
  ////    cnUnknown: Result := 'Unknown';
  //    cnTerminalSymbol: Result := 'TerminalSymbol';
  //    cnPragma: Result := 'Pragma';
  //    cnNonterminalSymbol: Result := 'NonterminalSymbol';
  //    cnCharClass: Result := 'CharClass';
  //    cnSingleChar: Result := 'SingleChar';
  //    cnWeakTerminalSymbol: Result := 'WeakTerminalSymbol';
  //    cnSymbolANY: Result := 'SymbolANY';
  //    cnEmptyAlternative: Result := 'EmptyAlternative';
  //    cnSymbolSYNC: Result := 'SymbolSYNC';
  //    cnSemanticAction: Result := 'SemanticAction';
  //    cnAlternative: Result := 'Alternative';
  //    cnIteration: Result := 'Iteration';
  //    cnOption: Result := 'Option';
  //    cnNoSym: Result := 'NoSym';
  //    cnEndOfFile: Result := 'EndOfFile';
  //  end;// Case
end;

procedure TpbCocoTable.SetTargetDirectory(const Value: string);
begin
  FTargetDirectory := Value;
end;

procedure TpbCocoTable.SethasTargetDirectory(const Value: boolean);
begin
  FhasTargetDirectory := Value;
end;

procedure TpbCocoTable.SethasErrorsSource(const Value: boolean);
begin
  FhasErrorsSource := Value;
end;

procedure TpbCocoTable.SetErrorsSource(const Value: string);
begin
  FErrorsSource := Value;
end;

end.
